--Quark 5/07
--developed for use with L and D
--converts an arbitrarily-long sequence of any of 4 chars to a compressed form
--and converts such a compressed-form sequence back to its original form and length
--compacts to 25% of original size

integer fn
sequence s1,s2,powersOfTwo

procedure PowersOfTwo(integer maxPower) 
    --set up sequence of the powers of two
    powersOfTwo={}
    for i = maxPower to 0 by -1 do
        powersOfTwo&=power(2,i)
    end for
end procedure

function BlockToCode(sequence b,integer bb)
    if bb=b[1] then
        return "00"
    elsif bb=b[2] then
        return "01"
    elsif bb=b[3] then
        return "10"
    elsif bb=b[4] then
        return "11"
    else
        return "XX"--error
    end if  
end function

function CodeToNum(sequence s)
    --s is a sequence of 4 groups of "00" or "01" or "10" or "11" (codes 4 values)
    integer ptr
    atom x
    sequence s1
    x=0
    ptr=0
    s1={}
    for i=1 to length(s) do
        s1=s[i]
        if s1[1]='1' then
            x+=powersOfTwo[i+ptr]
        end if      
        if s1[2]='1' then
            x+=powersOfTwo[i+ptr+1]
        end if
        ptr+=1
    end for
    return x
end function

function NumToCode(atom x)
    --a coded atom is restored to a sequence of 4 values: "00" or "01" or "10" or "11"
    integer ptr
    sequence s
    s=repeat("00",4)
    ptr=0
    for i = 1 to length(s) do
        if x>=powersOfTwo[i+ptr] then
            s[i][1]='1'
            x-=powersOfTwo[i+ptr]
        end if
        if x>=powersOfTwo[i+ptr+1] then
            s[i][2]='1'
            x-=powersOfTwo[i+ptr+1]
        end if
        ptr+=1
    end for
    return s
end function

function CodeToBlock(sequence b,sequence s)
    if equal(s,"00") then
        return b[1]
    elsif equal(s,"01") then
        return b[2]
    elsif equal(s,"10") then
        return b[3]
    elsif equal(s,"11") then
        return b[4]
    else
        return 'X'--error
    end if  
end function

global function CharFourToOne(sequence b,sequence s)
    --s is a long sequence of any of four specific chars (e.g."ABCD")
    sequence s1,s2,block
    PowersOfTwo(7)--set up array of powers of 2
    s1={}
     for i=1 to length(s) by 4 do
        block=s[i..i+3]
        s2={}
        for j=1 to length(block) do
            s2=append(s2,BlockToCode("WOPA",block[j]))-- e.g. {{00},{01},{10},{11}}
        end for
        s1&=CodeToNum(s2)
    end for
    return s1 -- array of numbers each coding four numbers
end function

global function CharOneToFour(sequence b,sequence s)
    --s is a long sequence of coded numbers, each containing 4 numbers
    sequence s1,s2
    PowersOfTwo(7)--set up array of powers of 2
    s1={}
    for i=1 to length(s) do
        s2=NumToCode(s[i]) --convert from number to a sequence of 4 values: "00" or "01" or "10" or "11"
        for j=1 to length(s2) do
            s1&=CodeToBlock(b,s2[j]) -- convert the 4 values to the original characters
        end for
    end for
    return s1 --the restored sequence of any of four specific chars (e.g."ABCD")
end function
